// GBA MMU emulator
// Mic, 2004

#include <windows.h>
#include <commctrl.h>
#include <string.h>
#include "zombie.h"
#include <stdio.h>

#define JUMP 0

#define NR10	0x60
#define NR11	0x62
#define NR12	0x63
#define NR13	0x64
#define NR14	0x65
#define NR21	0x68
#define NR22	0x69
#define NR23	0x6C
#define NR24	0x6D
#define NR30	0x70
#define NR31	0x72
#define NR32	0x73
#define NR33	0x74
#define NR34	0x75
#define NR41	0x78
#define NR42	0x79
#define NR43	0x7C
#define NR44	0x7D
#define NR50	0x80
#define NR51	0x81
#define NR52	0x84

#define TM0D	0x100
#define TM0CNT	0x102
#define TM1D	0x104
#define TM1CNT	0x106
#define TM2D	0x108
#define TM2CNT	0x10A
#define TM3D	0x10C
#define TM3CNT	0x10E
#define P1		0x130
#define P1CNT	0x132
#define IE		0x200
#define IF_		0x202
#define IME		0x208


typedef struct
{
	int sad,
		dad,
		cnt_l,
		cnt_h;
}dma_channel;

typedef struct
{
	unsigned int tstart,tcurr;
	bool enabled;
}timer_channel;

const unsigned char prescalar[4] = {0,6,8,10};

HINSTANCE hInst;
HANDLE hout;
unsigned char *ROM,*SYSROM,*IWRAM,*EXRAM,*cart,*VRAM,*IOREGS,*mmu_memory;
unsigned char *iomap,**mmap,*pageTable[16];
unsigned int pageLength[16],pageMask[16];
unsigned short int keys;
int *cpu_regs;
int romOffset,romSize;
int iwramOffset,iwramSize;
int exramOffset,exramSize;
int vramOffset,vramSize;
char mmuName[] = "GBA MMU";
int dummy;
unsigned int *cycle;
cpu_irq_callback cpu_irq;
apu_write_callback apu_write;
zombie_callback zombie;
char buffer[80];
FILE *log;
dma_channel dma[4];
timer_channel timer[4];

unsigned int sysromcode_0[8] = {
	0,			// 00000000
	0,			// 00000004
	0,			// 00000008
	0,			// 0000000C
	0,			// 00000010
	0,			// 00000014
	0xEA000042,	// 00000018: B $128
	0			// 0000001C
};

unsigned int sysromcode_128[6] = {
	0xE92D500F,	// 00000128: STMFD R13!,{R0-R3,R12,R14}
	0xE3A00301,	// 0000012C: MOV R0,#$04000000
	0xE28FE000,	// 00000130: ADD R14,R15,#0
	0xE510F004,	// 00000134: LDR R15,[R0,-4]
	0xE8BD500F,	// 00000138: LDMFD R13!,{R0-R3,R12,R14}
	0xE25EF004	// 0000013C: SUBS R15,R14,#4
};




BOOL CALLBACK DlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam);



bool __stdcall DllMain(HINSTANCE hDLL,DWORD reason,LPVOID unused)
{
	if (reason==DLL_PROCESS_ATTACH)
		hInst = hDLL;

	return true;
}



bool __stdcall mmu_init(zombie_callback commlink)
{
	int i;

	zombie = commlink;

	romOffset = 0x8000000;
	romSize = 0x8000000;

	exramOffset = 0x2000000;
	exramSize = 0x40000;

	iwramOffset = 0x3000000;
	iwramSize = 0x8000;

	vramOffset = 0x6000000;
	vramSize = 0x40000;

	mmu_memory = (unsigned char*)LocalAlloc(LPTR,iwramSize+exramSize+0x4000);
	IWRAM = mmu_memory;
	EXRAM = IWRAM + iwramSize;
	SYSROM = EXRAM + exramSize;
	//IOREGS = (unsigned char*)LocalAlloc(LPTR,0x400);

	pageTable[3] = IWRAM;
	pageTable[2] = EXRAM;
	pageTable[0] = SYSROM;

	pageLength[3] = 0x8000;
	pageLength[2] = 0x40000;
	pageLength[0] = 0x4000;

	for (i=0; i<4; i++)
	{
		dma[i].sad = 0xB0 + (i*0xC);
		dma[i].dad = 0xB4 + (i*0xC);
		dma[i].cnt_l = 0xB8 + (i*0xC);
		dma[i].cnt_h = 0xBA + (i*0xC);

		timer[i].enabled = false;
		timer[i].tcurr = 0;
		timer[i].tstart = 0;
	}

	/*iomap = new unsigned char[0x10000];

	for (i=0; i<0x10000; i++)
		iomap[i] = 0xFF;*/

	//log = fopen("mmulog.txt","wb");

	hout = GetStdHandle(STD_OUTPUT_HANDLE);
	return true;

}


bool __stdcall mmu_reset()
{
	int i;

	//cpu_inpb = (cpu_inpb_callback)zombie(ASK,CPU_INPB_CALLBACK,0,0);
	//cpu_outpb = (cpu_outpb_callback)zombie(ASK,CPU_OUTPB_CALLBACK,0,0);

	cpu_irq = (cpu_irq_callback)zombie(ASK,CPU_IRQ_CALLBACK,0,0);
	apu_write = (apu_write_callback)zombie(ASK,APU_WRITE_CALLBACK,0,0);
	cpu_regs = (int*)zombie(ASK,CPU_REG,0,0);

	cart = (unsigned char*)zombie(ASK,ZOMBIE_ROM_BASE,0,0);
	ROM = cart;

	zombie(TELL,GPU_VBLANK_ACTION,(void*)JUMP,(void*)0x38);
	zombie(TELL,GPU_HBLANK_ACTION,(void*)JUMP,(void*)0x38);

	zombie(TELL,CPU_REG_PC,(void*)0x8000000,0);
	VRAM = (unsigned char*)zombie(ASK,GPU_MEMORY,0,0);
	IOREGS = &VRAM[0x1D000];

	IOREGS[0x130] = 0xFF;
	keys = 0x3FF;

	pageTable[0x8] = ROM;
	pageTable[0xA] = ROM;
	pageTable[0xC] = ROM;
	pageTable[7] = VRAM+0x20000;
	pageTable[6] = VRAM;
	pageTable[5] = VRAM+0x1C000;
	pageTable[4] = IOREGS;

	pageLength[8] = (unsigned int)zombie(ASK,ZOMBIE_ROM_SIZE,0,0);
	pageLength[7] = 0x400;
	pageLength[6] = 0x18000;
	pageLength[5] = 0x400;
	pageLength[4] = 0x400;

	for (i=0; i<16; i++)
		pageMask[i] = pageLength[i] - 1;

	// Align the ROM size mask to the nearest 2^n boundary
	i = 0;
	while (true)
	{
		if (((1<<i)<=pageMask[8]) && ((1<<(i+1))>pageMask[8]))
			break;
		i++;
	}
	pageMask[8] = (1<<(i+1)) - 1;

	// Clear IWRAM
	memset(IWRAM,0,iwramSize);

	memcpy(SYSROM,sysromcode_0,8*4);
	memcpy(SYSROM+0x128,sysromcode_128,6*4);

	srand(123);

	return true;
}


void __stdcall mmu_close()
{
	//fclose(log);
	LocalFree(mmu_memory);
}



int __stdcall mmu_config(HWND hwnd)
{
	return 0; /*DialogBoxParam(hInst,
            MAKEINTRESOURCE(IDD_MMU_CONFIG), hwnd,
            DlgProc, NULL);*/
}




unsigned char* __stdcall mmu_translate_address(unsigned int address)
{
	return pageTable[address>>24] + (address&0xFFFFFF);
}



void dma_transfer(void *src,void *dest,int cnt)
{
	memcpy(dest,src,cnt);
}



unsigned short int __stdcall mmu_read_word(unsigned int);
void __stdcall mmu_write_word(unsigned int,int);


void dma_setup(int channel,bool forceDMA)
{
	int i,cnt;
	unsigned char sadcnt;
	unsigned short int cnts;
	unsigned int src,dest,srcend,destend;

	if ( ((IOREGS[dma[channel].cnt_h+1] & 0xB0) == 0x80) ||
		 (((IOREGS[dma[channel].cnt_h+1] & 0xB0) >= 0x80) && forceDMA) )
	{
		cnts = *(unsigned short*)(IOREGS + dma[channel].cnt_l);
		cnt = cnts;
		cnt <<= ((IOREGS[dma[channel].cnt_h+1] & 4) == 4) ? 2 : 1;

		src = (*(unsigned int*)(IOREGS + dma[channel].sad)) & 0xFFFFFFF;
		dest = (*(unsigned int*)(IOREGS + dma[channel].dad)) & 0xFFFFFFF;

		srcend = (src+cnt)&0xFFFFFF;
		if (srcend >= pageLength[src>>24])
			cnt -= (srcend-pageLength[src>>24]);
		if (cnt<=0) return;
		destend = (dest+cnt)&0xFFFFFF;
		if (destend >= pageLength[dest>>24])
			cnt -= (destend-pageLength[dest>>24]);
		if (cnt<=0) return;

		/*wsprintf(buffer,"DMA:ing %d bytes from %08x to %08x (%04x)\n",
			cnt,src,dest,cnts);
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);*/

		dma_transfer((void*)mmu_translate_address(src),(void*)mmu_translate_address(dest),cnt);

		//for (i=cnt; i>0; i-=2)
		//	mmu_write_word(dest+i-2,mmu_read_word(src+i-2));

		switch (IOREGS[dma[channel].cnt_h] & 0x60)
		{
			case 0x00:
				dest += cnt;
				break;
			case 0x20:
				dest -= cnt;
				break;
			default:
				break;
		}
		*(unsigned int*)(IOREGS + dma[channel].dad) = dest;

		sadcnt = IOREGS[dma[channel].cnt_h] & 0x80;
		sadcnt |= IOREGS[dma[channel].cnt_h+1] & 0x01;
		switch (sadcnt)
		{
			case 0x00:
				src += cnt;
				break;
			case 0x80:
				src -= cnt;
				break;
			default:
				break;
		}
		*(unsigned int*)(IOREGS + dma[channel].sad) = src;

		// Reset the DMA enable bit unless repeat mode is specified
		if ((IOREGS[dma[channel].cnt_h+1] & 0x02) == 0)
			IOREGS[dma[channel].cnt_h+1] &= 0x7F;
	}
}



void timer_setup(int channel)
{
	timer[channel].tstart = *cycle;

	timer[channel].enabled = false;
	if (IOREGS[TM0CNT+channel*4]&0x80)
		timer[channel].enabled = true;
}




void __stdcall mmu_irq(int mask)
{
	int i;

	if (mask & 2)
		for (i=1; i<4; i++)
			if ((IOREGS[dma[i].cnt_h+1] & 0xB0) == 0xA0)
				dma_setup(i,true);

	if (mask & 1)
		for (i=1; i<4; i++)
			if ((IOREGS[dma[i].cnt_h+1] & 0xB0) == 0x90)
				dma_setup(i,true);
}



unsigned char __stdcall mmu_read_byte(unsigned int address)
{
	unsigned int page = address>>24,offset; //=address&0xFFFFFF;

	if (pageTable[page])
		//if (pageLength[page]>offset)
			return *(unsigned char*)(pageTable[page] + (address&pageMask[page]));

	return rand()&0xFF;
}


unsigned short int __stdcall mmu_read_word(unsigned int address)
{
	unsigned int page = address>>24,offset=address&0xFFFFFF;

	if (page==4)
		switch (offset)
		{
			case P1:
				return keys;
			case TM2D:
				if (timer[2].enabled)
				{
					timer[2].tcurr = *cycle;
					if (timer[2].tcurr < timer[2].tstart)
						timer[2].tcurr -= (timer[2].tstart-0x40000000);
					else
						timer[2].tcurr -= timer[2].tstart;
					timer[2].tcurr >>= prescalar[IOREGS[TM2CNT]&3];
				}
				return timer[3].tcurr &= 0xFFFF;
			case TM3D:
				if (timer[3].enabled)
				{
					timer[3].tcurr = *cycle;
					if (timer[3].tcurr < timer[3].tstart)
						timer[3].tcurr -= (timer[3].tstart-0x40000000);
					else
						timer[3].tcurr -= timer[3].tstart;
					timer[3].tcurr >>= prescalar[IOREGS[TM3CNT]&3];
				}
				return timer[3].tcurr &= 0xFFFF;

			case 0x10:
			case 0x12:
			case 0x14:
			case 0x16:
			case 0x18:
			case 0x1A:
			case 0x1C:
			case 0x1E:
				page = cpu_regs[15];
				offset = page&0xFFFFFF;
				return *(unsigned short int*)(pageTable[page>>24] + offset);

			default:
				return *(unsigned short int*)(pageTable[page] + offset);
		}

	if (pageTable[page])
		return *(unsigned short int*)(pageTable[page] + offset);
	return rand()&0xFFFF;
}


int __stdcall mmu_read_dword(unsigned int address)
{
	unsigned int page = (address>>24)&0x0F,offset=address&0xFFFFFF;

	if (page==8)
	{
		while (offset >= pageLength[page])
			offset -= pageLength[page];
		return *(int*)(pageTable[page] + offset);
	}
	if (pageTable[page])
		//if (pageLength[page]>offset)
			return *(int*)(pageTable[page] + (address&pageMask[page]));

	return rand()&0xFFFFFFFF;
}



void __stdcall mmu_write_byte(unsigned int address,int byte)
{
	unsigned int page = (address>>24)&0x0F,offset=address&0xFFFFFF;

	if (pageTable[page])
		if (pageLength[page]>offset)
			*(char*)(pageTable[page] + offset) = (char)byte;

	if ((page==4)&&(offset>=NR10)&&(offset<=NR52))
		apu_write(offset,byte&0xFF);

	/*if (page == 6)
		if (offset&1)
			*(char*)(pageTable[page] + offset-1) = (char)byte;
		else
			*(char*)(pageTable[page] + offset+1) = (char)byte;*/
}



void __stdcall mmu_write_word(unsigned int address,int word)
{
	unsigned int page = address>>24,offset=address&0xFFFFFF;

	if (pageTable[page])
		if (pageLength[page]>offset)
			*(unsigned short int*)(pageTable[page] + offset) = (unsigned short int)word;


	if (page == 4)
	{
		if ((offset>=NR10)&&(offset<=NR52))
		{
			apu_write(offset,word&0xFF);
			apu_write(offset+1,(word>>8)&0xFF);
		}

		switch (offset)
		{
			case 0xC6:
				dma_setup(1,false);
				break;
			case 0xD2:
				dma_setup(2,false);
				break;
			case 0xDE:
				dma_setup(3,false);
				break;
			case TM0D:
				timer[0].tcurr = word&0xFFFF;
				break;
			case TM1D:
				timer[1].tcurr = word&0xFFFF;
				break;
			case TM2D:
				timer[2].tcurr = word&0xFFFF;
				break;
			case TM3D:
				timer[3].tcurr = word&0xFFFF;
				break;
			case TM0CNT:
				timer_setup(0);
				break;
			case TM1CNT:
				timer_setup(1);
				break;
			case TM2CNT:
				timer_setup(2);
				break;
			case TM3CNT:
				timer_setup(3);
				break;
			//case IF_:
			default:
				break;
		}
	}
}



void __stdcall mmu_write_dword(unsigned int address,int dwrd)
{
	unsigned int page = address>>24,offset=address&0xFFFFFF;

	if (pageTable[page])
		if (pageLength[page]>offset)
			*(unsigned int*)(pageTable[page] + offset) = (unsigned int)dwrd;

	if (page == 4)
	{
		switch (address & 0xFFFFFF)
		{
			case 0xC4:
				dma_setup(1,false);
				break;
			case 0xC6:
				dma_setup(1,false);
				break;
			case 0xD0:
				dma_setup(2,false);
				break;
			case 0xD2:
				dma_setup(2,false);
				break;
			/*case 0xD4:
				wsprintf(buffer,"DM3SAD = %08x\n",dwrd);
				WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
				break;*/
			case 0xDC:
				dma_setup(3,false);
				break;
			case 0xDE:
				dma_setup(3,false);
				break;
			default:
				break;
		}
	}
}



int __stdcall mmu_get(int what)
{
	switch (what)
	{
		case MMU_READ_BYTE_CALLBACK:
			return (int)mmu_read_byte;
		case MMU_READ_WORD_CALLBACK:
			return (int)mmu_read_word;
		case MMU_READ_DWORD_CALLBACK:
			return (int)mmu_read_dword;
		case MMU_WRITE_BYTE_CALLBACK:
			return (int)mmu_write_byte;
		case MMU_WRITE_WORD_CALLBACK:
			return (int)mmu_write_word;
		case MMU_WRITE_DWORD_CALLBACK:
			return (int)mmu_write_dword;
		case MMU_TRANSLATE_ADDRESS_CALLBACK:
			return (int)mmu_translate_address;
		case MMU_IRQ_CALLBACK:
			return (int)mmu_irq;
		case PLUGIN_NAME:
			return (int)&mmuName;
		case 0x700:
			return (int)mmu_translate_address((unsigned int)0x8000288);
		default:
			break;
	}
	return 0;
}


void __stdcall mmu_set(int what,void* param1,void* param2)
{
	static char buffer[64];

	switch (what)
	{
		case CPU_CYCLE_PTR:
			cycle = (unsigned int *)param1;
			break;

		case ZOMBIE_KEYSTATE:
			if ((int)param2==1)
			{
				switch ((int)param1)
				{
					case 38:	// Up
						keys &= 0x3BF;
						break;
					case 40:
						keys &= 0x37F;
						break;
					case 37:	// Left
						keys &= 0x3DF;
						break;
					case 39:
						keys &= 0x3EF;
						break;
					case 13:
						keys &= 0x3F7;
						break;
					case 'X':
						keys &= 0x3FE;
						break;
					case 'Z':
						keys &= 0x3FD;
						break;
					case 'L':
						keys &= 0x1FF;
						break;
					case 'R':
						keys &= 0x2FF;
						break;
					case ' ':
						keys &= 0x3FB;
						break;
				}

			} else
			{
				switch ((int)param1)
				{
					case 38:	// Up
						keys |= 0x40;
						break;
					case 40:
						keys |= 0x80;
						break;
					case 37:	// Left
						keys |= 0x20;
						break;
					case 39:
						keys |= 0x10;
						break;
					case 13:
						keys |= 0x08;
						break;
					case 'X':
						keys |= 0x01;
						break;
					case 'Z':
						keys |= 0x02;
						break;
					case 'L':
						keys |= 0x200;
						break;
					case 'R':
						keys |= 0x100;
						break;
					case ' ':
						keys |= 0x04;
						break;
				}
			}
			if (!IOREGS) return;

			if (IOREGS[IME]&1)
				if (IOREGS[IE+1]&0x20)
					if (IOREGS[P1CNT+1]&0x40)
						if (IOREGS[P1CNT+1]&0x80)
							if (keys&IOREGS[P1CNT])
							{
								IOREGS[0x201] |= 0x20;
								cpu_irq(0x00000018);
							}
							else if ((keys&IOREGS[P1CNT])!=IOREGS[P1CNT])
							{
								IOREGS[0x201] |= 0x20;
								cpu_irq(0x00000018);
								//keys = keys; // key interrupt
							}
			break;
		default:
			break;
	}
}


